"""A package for matrix completion with column subset selection."""

__version__ = "0.1.0"


from .csmc import CSMC
from .mc.nn_completion import NuclearNormMin
from .mc.soft_impute import SoftImpute



"""CSMC logic."""

from abc import abstractmethod
from collections.abc import Callable
from enum import Enum
from typing import Generic

import numpy as np
import torch
from torch import Tensor

from csmc.settings import T, LOGGER, UNSUPPORTED_MSG
from csmc.css import uniform
from csmc.mc.nn_completion import NuclearNormMin
from csmc.transform import dls



class FillMethod(Enum):
    """Options for the initial imputation."""

    ZERO = "zero"
    MEAN = "mean"
    MEDIAN = "median"
    MIN = "min"


class CSMCBase(Generic[T]):
    """Base class for CSMC model."""

    def __init__(self, X: T, col_number: int, col_select: Callable = uniform,
                 transform: Callable = dls, solver: Callable = NuclearNormMin,
                 threshold: float = 0, fill_method: FillMethod = FillMethod.ZERO,
                 lambda_: float = 0, max_rank: int | None = None) -> None:
        self.col_number = col_number
        self.col_select = col_select
        self._transform = transform
        self.threshold = threshold
        self.fill_method = fill_method
        self.solver = solver
        self.lambda_ = lambda_
        self.max_rank = max_rank
        self.C_incomplete = None

    def fit_transform(self, X: T) -> T:
        """Complete matrix with CSMC."""
        X_tmp = self._copy(X)
        missing_mask = self._missing_mask(X_tmp)
        self._prepare(X_tmp, missing_mask)
        if self.C_incomplete is None:
            cols_indices = self.col_select(X_tmp, self.col_number)
            self.cols_indices = cols_indices
            C_incomplete = X_tmp[:, cols_indices]
            cols_missing = missing_mask[:, cols_indices]
        else:
            cols_indices = self.cols_indices
            C_incomplete = self.C_incomplete
            cols_missing = self.cols_missing
        C_filled = self.fill_columns(C_incomplete, cols_missing)
        LOGGER.debug("Column submatrix filled.")
        return self.transform(X, C_filled, cols_indices, ~missing_mask)

    def fill_columns(self, C_incomplete: T, missing_mask: T) -> T:
        """Complete column submatrix."""
        solver = self.solver(C_incomplete, lambda_=self.lambda_,
                             max_rank=self.max_rank) if self.lambda_ else self.solver(C_incomplete)
        return solver.fit_transform(C_incomplete, missing_mask)

    def _fill_columns_with_fn(self, X: T, missing_mask: T, col_fn: Callable) -> None:
        for col_idx in range(X.shape[1]):
            missing_col = missing_mask[:, col_idx]
            n_missing = missing_col.sum()
            if n_missing == 0:
                continue
            col_data = X[:, col_idx]
            fill_values = col_fn(col_data)
            if np.all(np.isnan(fill_values)):
                fill_values = 0
            X[missing_col, col_idx] = fill_values

    def transform(self, X_org: T, C_filled: T, cols_indices: T,
                  ok_mask: T) -> T:
        """Solve the least squares problem."""
        X_filled = self._copy(X_org)
        for i, ci in enumerate(cols_indices):
            X_filled[:, ci] = C_filled[:, i]
        X_filled[ok_mask] = X_org[ok_mask]
        out = self._transform(X_filled, ok_mask, C_filled)
        out[ok_mask] = X_org[ok_mask]
        return out

    @abstractmethod
    def _copy(self, X: T) -> T:
        pass

    @abstractmethod
    def _missing_mask(self, X: T) -> T:
        pass

    @abstractmethod
    def _prepare(self, X: T, missing_mask: T) -> None:
        pass


class CSMC_N(CSMCBase[np.ndarray]):
    """Class for CSMC for numpy arrays."""

    def _copy(self, X: np.ndarray) -> np.ndarray:
        return np.copy(X)

    def _missing_mask(self, X: np.ndarray) -> np.ndarray:
        return np.isnan(X)

    def _prepare(self, X: np.ndarray, missing_mask: np.ndarray) -> None:
        if self.fill_method == FillMethod.ZERO:
            X[missing_mask] = 0
        elif self.fill_method == FillMethod.MEAN:
            self._fill_columns_with_fn(X, missing_mask, np.nanmean)
        elif self.fill_method == FillMethod.MEDIAN:
            self._fill_columns_with_fn(X, missing_mask, np.nanmedian)
        elif self.fill_method == FillMethod.MIN:
            self._fill_columns_with_fn(X, missing_mask, np.nanmin)


class CSMC_T(CSMCBase[Tensor]):
    """Class for CSMC for torch tensors."""

    def __init__(self, X: Tensor, **kwargs) -> None:
        super().__init__(X, **kwargs)
        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    def _copy(self, X: Tensor) -> Tensor:
        return torch.clone(X)

    def _missing_mask(self, X: Tensor) -> Tensor:
        return torch.isnan(X)

    def _prepare(self, X: Tensor, missing_mask: Tensor) -> None:
        X[missing_mask] = 0


class CSMC:
    """Class for CSMC."""

    def __new__(cls, X: T, *args, **kwargs):
        """Create CSMC object based on the type of X."""
        if isinstance(X, np.ndarray):
            return CSMC_N(X, *args, **kwargs)
        elif isinstance(X, torch.Tensor):
            return CSMC_T(X, *args, **kwargs)
        else:
            raise TypeError(UNSUPPORTED_MSG)



"""Matrix completion with nuclear norm minimization using SDP solvers."""
import cvxpy as cp
import numpy as np


class NuclearNormMin:
    """Class for completing matrix using nuclear norm minimization as SDP."""

    def __init__(self, M_incomplete: np.ndarray, solver: str = "SCS") -> None:
        self.solver = solver

    def fit_transform(self, M_incomplete: np.ndarray, missing_mask: np.ndarray) -> np.ndarray:
        """Matrix completion logic."""
        return self.nn_complete(M_incomplete, missing_mask)

    def nn_complete(self, M_incomplete: np.ndarray, missing_mask: np.ndarray | None) -> np.ndarray:
        """Fill M_incomplete with nuclear norm minimization."""
        if missing_mask is None:
            missing_mask = np.isnan(M_incomplete)
        M_incomplete[missing_mask] = 0
        M_filled = cp.Variable(M_incomplete.shape)
        prob = cp.Problem(cp.Minimize(cp.norm(M_filled, p="nuc")),
                          [cp.multiply(~missing_mask, M_filled) == cp.multiply(~missing_mask, M_incomplete)])
        prob.solve(solver=cp.SCS, verbose=False)
        return M_filled.value



import logging
import sys
from typing import TypeVar

from numpy import ndarray
from torch import Tensor
import torch
LOGGER = logging.getLogger("__csmc__")
LOGGER.propagate = False
LOGGER.setLevel(logging.INFO)
_stream_handler = logging.StreamHandler(sys.stdout)
_stream_handler.setLevel(logging.INFO)
_formatter = logging.Formatter(
    fmt="(CSMC) %(asctime)s: %(message)s", datefmt="%b %d %I:%M:%S %p"
)
_stream_handler.setFormatter(_formatter)
LOGGER.addHandler(_stream_handler)

T = TypeVar("T", ndarray, Tensor)

UNSUPPORTED_MSG ="Unsupported data type for X."

torch.set_num_threads(8)


"""The second stage of CSMC methods."""

import math

import cvxpy as cp
import numpy as np
import torch
from scipy.linalg import lstsq
from torch import Tensor
from torch.autograd.variable import Variable

from csmc.settings import T


# @numba.njit(parallel=True)
# def _dls(X: np.ndarray, ok_mask: np.ndarray, C: np.ndarray, Y: np.ndarray, n: int) -> None:
#     """Solve the least squares for each column of array X."""
#
#     for i in numba.prange(n):
#         mask_i = ok_mask[:, i]
#
#         #
#         # # Using numpy's pinv and matrix multiplication for solving the least squares problem
#         Y[i, :] = np.linalg.pinv(C[mask_i]) @ X[mask_i, i]


def _dls(X: np.ndarray, ok_mask: np.ndarray, C: np.ndarray, Y: np.ndarray, n: int) -> None:
    """Solve the least squares for each column of array X."""

    for i in range(n):
        mask_i = ok_mask[:, i]
        Y[:, i] = lstsq(C[mask_i], X[mask_i, i], lapack_driver="gelsy")[0]


def _dls_torch(X: T, ok_mask: T, C: T, Y: T, n: int, cuda_support: bool) -> None:
    for i in range(n):
        mask_i = ok_mask[:, i]
        if cuda_support:
            try:
                Y[i, :] = torch.linalg.lstsq(C[mask_i], X[mask_i, i], driver="gels")[0]
            except RuntimeError:
                print("Falling back")
                Y[i, :] = torch.linalg.pinv(C[mask_i]) @ X[mask_i, i]
        else:
            try:
                Y[:, i] = torch.linalg.lstsq(C[mask_i], X[mask_i, i], driver="gelsy")[0]
            except RuntimeError:
                print("Falling back")
                Y[:, i] = torch.linalg.lstsq(C[mask_i], X[mask_i, i], driver="gelsd")[0]


def dls(X: T, ok_mask: T, C: T) -> T:
    """Solve direct least squares."""
    m, n = X.shape
    _, k = C.shape
    if isinstance(X, np.ndarray):
        Y = np.empty((k, n), dtype=X.dtype)
        _dls(X, ok_mask, C, Y, n)
    else:
        cuda_support = torch.cuda.is_available()
        device = torch.device("cuda") if cuda_support else torch.device("cpu")
        Y = torch.zeros((k, n), dtype=torch.float32, device=device)
        _dls_torch(X, ok_mask, C, Y, n, cuda_support)
    return C @ Y


def ls_convex(X: T, ok_mask: T, C: T) -> T:
    """Solve the least squares with cvxpy optimizer."""
    m, n = X.shape
    _, k = C.shape
    Y = cp.Variable((k, n))
    X[~ok_mask] = 0
    obj = cp.sum_squares(cp.multiply(ok_mask, X) - cp.multiply(ok_mask, C @ Y))
    prob = cp.Problem(cp.Minimize(obj))
    prob.solve(solver=cp.SCS, use_indirect=False)
    return C @ Y.value


def sgrad_torch(X: Tensor, ok_mask: Tensor, C: Tensor, max_iters: int = 1000, lr: float | None = None) -> Tensor:
    """Solve the least squares with stochastic gradient descent."""
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    m, n = X.shape
    _, k = C.shape
    if lr is None:
        lr = 1 / math.sqrt(n)

    Y = Variable(torch.empty((k, n), dtype=torch.float32, device=device), requires_grad=True)
    Y.data.normal_(std=0.01)
    opt = torch.optim.Adam([Y], lr=lr)
    X[~ok_mask] = 0

    def loss():
        return torch.mean(torch.mul(ok_mask, (C.mm(Y) - X)) ** 2)

    for _ in range(max_iters):
        opt.zero_grad()
        loss_ = loss()
        loss_.backward()
        opt.step()
    return C @ Y



"""Evaluation methods."""
import math

import numpy as np
import torch
from torch import Tensor

from csmc.settings import T

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def approx_err(X: T, Y: T, numlib: str ="numpy") -> float:
    """Relative approximation error."""
    lib = np.linalg if numlib == "numpy" else torch
    return lib.linalg.norm(X - Y) / lib.linalg.norm(Y)


def approx_err_unknown(X: T, Y: T, missing_mask: T, numlib: str ="numpy") -> float:
    """Relative approximation over unknown entries."""
    lib = np.linalg if numlib == "numpy" else torch
    return lib.norm(X[missing_mask] - Y[missing_mask]) / lib.norm(Y[missing_mask])


def nmae_unknown(X: T, Y: T, ok_mask: T, numlib: str="numpy") -> float:
    """NMAE over unknown entries."""
    lib = np if numlib == "numpy" else torch
    return lib.sum(lib.abs(X[ok_mask] - Y[ok_mask])) / (lib.sum(ok_mask) * lib.max(Y[ok_mask]) - lib.min(Y[ok_mask]))


def approx_err_unknown_torch(X: Tensor, Y: Tensor, ok_mask: Tensor) -> float:
    """Relative approximation error for tensors."""
    return torch.norm(X[ok_mask] - Y[ok_mask]) / torch.norm(X[ok_mask])


def snr(X: T, Y: T, numlib: str="numpy") -> float:
    """Signal to noise ratio."""
    lib = np.linalg if numlib == "numpy" else torch
    return 20 * math.log10(lib.linalg.norm(Y) / lib.linalg.norm(X - Y))


def _round(X: T, numlib: str ="numpy") -> T:
    lib = np if numlib == "numpy" else torch
    X_r = lib.round(X * 2) / 2
    min_value = 0.5
    max_value = 5
    X_r[X_r < min_value] = min_value
    X_r[X_r > max_value] = min_value
    return X_r


def hits(X: T, Y: T, ok_mask: T, numlib: str="numpy") -> float:
    """Hits number."""
    X_r = _round(X, numlib)
    Y_r = _round(Y, numlib)
    lib = np if numlib == "numpy" else torch
    return lib.sum(X_r[ok_mask] == Y_r[ok_mask]) / lib.sum(ok_mask)



"""Matrix completion with Proximal Gradient Descent."""
from abc import abstractmethod
from typing import Any, Generic

import fbpca
import numpy as np
import torch
from numpy import floating, ndarray
from torch import Tensor

from csmc.settings import T, LOGGER, UNSUPPORTED_MSG


class SoftImputeBase(Generic[T]):
    """A class for completing matrix using Soft Impute algorithm, based on Proximal Gradient Descent method."""

    def __init__(self, X :T, lambda_: float = 10, threshold: float = 0.001, max_iter: int = 10000,
                 max_rank: int | None = None, numlib: str = "numpy", **kwargs) -> None:
        self.numlib = numlib
        self.lambda_ = lambda_
        self.threshold = threshold
        self.max_iter = max_iter
        self.max_rank = max_rank
        self.svd_randomized = True

    def shrinkage_operator(self, M: T, max_rank: int | None = None) -> tuple:
        """Apply Soft Thresholding to the matrix."""
        (U, s, V) = self._svd(M, max_rank)
        s_shrinked = self._shrink(s)
        rank = (s_shrinked > 0).sum()
        s_shrinked = s_shrinked[:rank]
        U_shrinked = U[:, :rank]
        V_shrinked = V[:rank, :]
        S_shrinked = self._diag(s_shrinked)
        M_shrinked = U_shrinked @ (S_shrinked @ V_shrinked)
        return M_shrinked, rank


    def solve(self, X: T, missing_mask: T,
              Z_init: T | None = None) -> T:
        """Perform Proximal Gradient Descent method."""
        Z_old = self._init(X.shape) if Z_init is None else Z_init
        ok_mask = ~missing_mask
        rank = self.max_rank
        for iter_ in range(self.max_iter):
            Z_old[ok_mask] = X[ok_mask]
            if rank:
                Z_new, rank = self.shrinkage_operator(Z_old, rank)
            else:
                Z_new, rank = self.shrinkage_operator(Z_old)
            if self._converged(Z_new, Z_old):
                LOGGER.debug(f"Converged after {iter_} iterations and rankd {rank} ")
                break
            Z_old = Z_new
        Z_new[ok_mask] = X[ok_mask]
        return Z_new

    def _converged(self, Z_new: T, Z_old: T) -> np.bool_:
        """Check convergence conditions."""
        return self._approx_err(Z_old, Z_new) < self.threshold

    def fit(self, M: T, missing_mask: T,
            Z_init: T | None = None) -> T:
        """Complete matrix."""
        return self.solve(M, missing_mask, Z_init=Z_init)

    def fit_transform(self, M: T, missing_mask: T, Z_init: T | None =None) -> T:
        """Complete matrix."""
        return self.fit(M, missing_mask, Z_init=Z_init)

    @abstractmethod
    def _svd(self, X: T, max_rank: int | None = None) -> tuple:
        pass

    @abstractmethod
    def _shrink(self, s: T) -> T:
        pass

    @abstractmethod
    def _diag(self, s: T) -> T:
        pass

    @abstractmethod
    def _init(self, shape: tuple) -> T:
        pass

    @abstractmethod
    def _ok_mask(self, X: T) -> T:
        pass

    @abstractmethod
    def _approx_err(self, X: T, Y: T) -> floating[Any]:
        pass


class SoftImpute_N(SoftImputeBase):
    """A numpy array interface."""

    def _svd(self, M: ndarray, max_rank: int | None = None) -> tuple:
        """Perform SVD."""
        if max_rank and self.svd_randomized:
            LOGGER.debug("Performing randomized SVD.")
            (U, s, V) = fbpca.pca(
                M,
                max_rank)

        else:
            LOGGER.debug("Performing full SVD.")
            (U, s, V) = np.linalg.svd(
                M,
                full_matrices=False,
                compute_uv=True)
            # (U, s, V) = fbpca.pca(
            #     M)
        return (U, s, V)

    def _shrink(self, s: ndarray) -> ndarray:
        """Shrink singular vector."""
        return np.maximum(s - self.lambda_, np.zeros_like(s))

    def _diag(self, s: ndarray) -> ndarray:
        """Diagonalize singular vector."""
        return np.diag(s)

    def _init(self, shape: tuple) -> ndarray:
        """Set initial point."""
        return np.zeros(shape)

    def _ok_mask(self, X: ndarray) -> ndarray:
        """Mask out missing values."""
        return ~np.isnan(X)

    def _approx_err(self, X: ndarray, Y: ndarray) -> floating[Any]:
        """Approximate error."""
        return np.linalg.norm(X - Y) / np.linalg.norm(X)


class SoftImpute_T(SoftImputeBase):
    """Torch tensor interface."""

    def __init__(self, *args,  **kwargs):
        super().__init__(*args, **kwargs)
        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    def _svd(self, M: Tensor, max_rank: int | None = None) -> tuple:
        """Perform SVD."""
        if max_rank and self.svd_randomized:
            (U, s, V) = torch.svd_lowrank(
                M,
                max_rank)
            V = V.T

        else:
            (U, s, V) = torch.linalg.svd(
                M,
                full_matrices=False)
        return (U, s, V)

    def _shrink(self, s: Tensor) -> Tensor:
        """Shrink singular vector."""
        return torch.maximum(s - self.lambda_, torch.zeros_like(s))

    def _diag(self, s: Tensor) -> Tensor:
        """Diagonalize singular vector."""
        return torch.diag(s)

    def _init(self, shape: tuple) -> Tensor:
        """Set initial point."""
        return torch.zeros(shape, dtype=torch.float32, device=self.device)

    def _ok_mask(self, X: Tensor) -> Tensor:
        """Mask out missing values."""
        return ~torch.isnan(X)

    def _approx_err(self, X: Tensor, Y: Tensor) -> floating[Any]:
        """Approximate error."""
        return torch.linalg.norm(X - Y) / torch.linalg.norm(X)


class SoftImpute:
    """Class for CSMC."""

    def __new__(cls, X: T, *args, **kwargs):
        """Create CSMC object based on the type of X."""
        if isinstance(X, np.ndarray):
            return SoftImpute_N(X, *args, **kwargs)
        elif isinstance(X, torch.Tensor):
            return SoftImpute_T(X, *args, **kwargs)
        else:
            raise TypeError(UNSUPPORTED_MSG)



"""Column Subset Selection methods."""

import numpy as np
import torch
from numpy import ndarray

from csmc.settings import T

"""
Column subset selection methods.
"""


def uniform(X: T, no_col: int) -> T:
    """Select columns according to uniform distribution."""
    if isinstance(X, ndarray):
        cols = np.random.default_rng().choice(X.shape[1], no_col, replace=False)
    else:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        cols = torch.randperm(X.shape[1], dtype=torch.int32, device=device)[:no_col]
    return cols



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['csmc', 'csmc.errors', 'csmc.mc']

package_data = \
{'': ['*']}

install_requires = \
['SciPy>=1.11.4,<2.0.0',
 'cvxpy>=1.4.1,<2.0.0',
 'fbpca>=1.0,<2.0',
 'numba>=0.58.1,<0.59.0',
 'threadpoolctl>=3.2.0,<4.0.0',
 'torch>=2.1.2,<3.0.0']

setup_kwargs = {
    'name': 'csmc',
    'version': '0.1.7',
    'description': 'Matrix completion with column subset selection.',
    'long_description': '# CSMC \n\nCSMC is a Python library for performing column subset selection in matrix completion tasks. It provides an implementation of the CSSMC method, which aims to complete missing entries in a matrix using a subset of columns.\n\nColumns Selected Matrix Completion (CSMC) is a two-stage approach for low-rank matrix recovery. In the first stage, CSMC samples columns of the input matrix  and recovers a smaller column submatrix.\nIn the second stage, it solves a least squares problem to reconstruct the whole matrix.\n\n<img src="resources/CSMC.png" alt="Alt text" width="400px" />\n\n\n## Installation\n\nYou can install CSMC using pip:\n\n```bash\npip install -i  csmc\n```\n\n## Usage\n1. Generate random data\n```python\nimport numpy as np\nimport random \n\nn_rows = 50\nn_cols = 250\nrank = 3\n\nx = np.random.default_rng().normal(size=(n_rows, rank)) \ny = np.random.default_rng().normal(size=(rank, n_cols)) \nM = np.dot(x, y)\n\nM_incomplete = np.copy(M)\nnum_missing_elements = int(0.7 * M.size)\nindices_to_zero = random.sample(range(M.size), k=num_missing_elements)\nrows, cols = np.unravel_index(indices_to_zero, M.shape)\nM_incomplete[rows, cols] = np.nan\n```\n\n2. Fill with CSNN algorithm\n```python\nfrom csmc import CSMC\nsolver = CSMC(M_incomplete, col_number=100)\nM_filled = solver.fit_transform(M_incomplete)\n```\n\n3. Fill with Nuclear Norm Minimization with SDP (NN algorithm)\n\n```python\nfrom csmc import NuclearNormMin\nsolver = NuclearNormMin(M_incomplete)\nM_filled = solver.fit_transform(M_incomplete, np.isnan(M_incomplete))\n```\n\n## Algorithms\n* `NuclearNormMin`: Matrix completion by SDP (NN algorithm) [Exact Matrix Completion via Convex Optimization](http://statweb.stanford.edu/~candes/papers/MatrixCompletion.pdf\n* `CSNN`: Matrix completion by CSNN\n* `PGD`: Nuclear norm minimization using Proximal Gradient Descent (PGD)  [Spectral Regularization Algorithms for Learning Large Incomplete Matrices](http://web.stanford.edu/~hastie/Papers/mazumder10a.pdf) by Mazumder et. al.\n* `CSPGD`: Matrix completion by CSPGD\n\n## Citation\n\nKrajewska, A., Niewiadomska-Szynkiewicz E. (2023). Matrix Completion with Column Subset Selection.\nKrajewska, A. (2023).  [Efficient matrix completion for data recovery in data-driven IT applications](https://www.e-bip.org.pl/upload/00998/37859/1088193-82224218.pdf). Systems Research Institute\nPolish Academy of Sciences.',
    'author': 'Antonina Krajewska',
    'author_email': 'antonina.krajewska@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ZAL-NASK/CSMC',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
